# Tool-Naranayaswami-Moynihan-Hodge model
# by Florian Pabst 2020


import time
import re
import os
import sys

import numpy as np
import matplotlib.pyplot as plt
import scipy.odr as odr
from scipy.integrate import quad
from scipy.optimize import curve_fit, fsolve

# Read data
dataName = sys.argv[1]
try:
    data = np.loadtxt(dataName,skiprows=0)
    print("Loading")
    dataTemp = data[:, 0]
    dataCp = data[:, 1]
except IndexError:
    print("File not found")
    exit()
     
dataOutName = os.path.splitext(str(dataName))[0] + "_Tfict+TNMH.dat"


# First step: Calculate fictive Temperature T_f(T) from Cp data, then dT_f/dT

# Find start and end point for glass Cp linear fit
coordsGl = [(150, -0.126), (160, -0.09)]
"""
print("Please click on the left limit for glass Cp linear fit")
fig, (ax1) = P.subplots(1)
ax1.set_xlabel('Temperature / K')
ax1.set_xlim(left=140)
ax1.set_xlim(right=200)
ax1.plot(dataTemp,dataCp, 'ro')
#ax1.legend()
def onclick(event):
  ix = event.xdata
  iy = event.ydata
  print(ix)

  global coordsGl
  coordsGl.append((ix, iy))
    
  if len(coordsGl) == 1:
    print('right limit')

  if len(coordsGl) == 2:
      fig.canvas.mpl_disconnect(cid)

  return coordsGl
cid = fig.canvas.mpl_connect('button_press_event', onclick)
P.show()
"""
print(coordsGl)

for i in range(0,len(dataTemp)):
    if dataTemp[i] < coordsGl[0][0]:
        fitLimitLeft_Gl = i
    if dataTemp[i] < coordsGl[1][0]:
        fitLimitRight_Gl = i

# Find start and end point for liquid Cp linear fit
coordsLq = [(190, 2.07), (200, 1.773)]
"""
print("Please click on the left limit for liquid Cp linear fit")
fig, (ax1) = P.subplots(1)
ax1.set_xlabel('Temperature / K')
ax1.set_xlim(left=180)
ax1.set_xlim(right=250)
ax1.plot(dataTemp,dataCp, 'ro')
#ax1.legend()
def onclick(event):
  ix = event.xdata
  iy = event.ydata
  print(ix)

  global coordsLq
  coordsLq.append((ix, iy))
    
  if len(coordsLq) == 1:
    print('right limit')

  if len(coordsLq) == 2:
      fig.canvas.mpl_disconnect(cid)

  return coordsLq
cid = fig.canvas.mpl_connect('button_press_event', onclick)
P.show()
#print(coordsLq)
"""

for i in range(0, len(dataTemp)):
    if dataTemp[i] < coordsLq[0][0]:
        fitLimitLeft_Lq = i
    if dataTemp[i] < coordsLq[1][0]:
        fitLimitRight_Lq = i


def slope(x, t, m):
    return t+x*m


# Determine glass Cp slope
res = curve_fit(slope, dataTemp[fitLimitLeft_Gl:fitLimitRight_Gl], dataCp[fitLimitLeft_Gl:fitLimitRight_Gl])
glassCpParam = res[0]

# Determine liquid Cp slope
res  = curve_fit(slope, dataTemp[fitLimitLeft_Lq:fitLimitRight_Lq], dataCp[fitLimitLeft_Lq:fitLimitRight_Lq])
liquidCpParam = res[0]

xLin = np.linspace(coordsGl[0][0], coordsLq[1][0], len(dataTemp))

CpInterpol = np.interp(xLin, dataTemp, dataCp)
"""
fig, (ax1) = P.subplots(1)
ax1.set_xlabel('Temperature / K')
ax1.set_xlim(left=coordsGl[0][0])
ax1.set_xlim(right=coordsLq[1][0])
ax1.set_ylim(bottom=coordsGl[0][1]-0.1)
ax1.set_ylim(top=coordsLq[1][1]+0.1)
ax1.plot(dataTemp,dataCp, 'ro')
ax1.plot(dataTemp,slope(dataTemp,glassCpParam[0],glassCpParam[1]), 'b-')
ax1.plot(dataTemp,slope(dataTemp,liquidCpParam[0],liquidCpParam[1]), 'b-')
ax1.plot(xLin,CpInterpol, 'g-')
P.show()
"""


# Calculate Fictive Temperature for all temperatures in range
intStart = fitLimitLeft_Gl
intStop = fitLimitRight_Lq

fictiveTemp = []
temperatures = []
for i,tempStart in enumerate(dataTemp):
    if fitLimitLeft_Gl < i < fitLimitRight_Lq:
        intStart = i

        integration2 = np.trapz((dataCp[intStart:intStop]-slope(dataTemp[intStart:intStop], glassCpParam[0], glassCpParam[1])), dataTemp[intStart:intStop])

        def integrand1(temp):
            return slope(temp, liquidCpParam[0], liquidCpParam[1])-slope(temp, glassCpParam[0], glassCpParam[1])

        def fictiveTempEq(x):
            return quad(integrand1, x, coordsLq[1][0])[0] - integration2

        ficTemp = fsolve(fictiveTempEq, np.array([160]))
        fictiveTemp.append(ficTemp[0])
        temperatures.append(tempStart)


# Determine limiting T_f, i.e. T_f at lowest temperatures
def constant(x, t):
    return t


res = curve_fit(constant, temperatures[0:20], fictiveTemp[0:20])
T_f_0 = res[0]
print(T_f_0[0])

# Calculate dTf/dT
dTfdT = []
temperaturesCut = []
for i in range(len(temperatures)-1):
    if (temperatures[i+1]-temperatures[i]) != 0:
        dTfdT.append((fictiveTemp[i+1]-fictiveTemp[i]) / (temperatures[i+1]-temperatures[i]))
        temperaturesCut.append(temperatures[i+1])

"""        
fig, (ax1) = P.subplots(1)
ax1.set_xlabel('Temperature / K')
ax1.set_ylabel('Fictive Temperature / K')
#ax1.set_xlim(left=coordsGl[0][0])
#ax1.set_xlim(right=coordsLq[1][0])
#ax1.set_ylim(bottom=coordsGl[0][1]-0.1)
#ax1.set_ylim(top=coordsLq[1][1]+0.1)
ax1.plot(temperatures,fictiveTemp, 'ro')
#ax1.plot(temperatures,(T_f_0+temperatures*0), 'b-')
ax1.axhline(y=T_f_0[0], color='b', linestyle='-')
#ax1.plot(temperatures[:-1],dTfdT, 'ro')
P.show()
"""

# Second step: Calcualte and Fit TNMH-model
try:
    rate = float(str((re.findall(r'\d+\.\d+K', dataName))[0])[:-1])
    print("\tRate:  %fK-min" % rate)
except IndexError:
    rate = int(input("\tPlease enter the heating rate:   "))
    print("\tRate:  %fK-min" %rate)


def tau_k(tau_0, deltaE, T_k, xx, T_f_km1):
    return tau_0 * np.exp(xx*deltaE/(R*T_k) + (1-xx)*deltaE/(R*T_f_km1))
    

exponents = []
taus = []
temps = [0]
T_f_ns = []
markovs = [coordsLq[1][0]]
R = 8.31

# TNMH Code (adapted from Badrinarayanan's PhD Thesis):
def tnmfunc2(xdata, tau_0, xx, beta, deltaE):
    delhR = deltaE/8.314
    T = []
    Tf = []
    t = []
    delt = []
    tau = []
    dttau = []
    delT = []
    step = len(xdata)
    for k in range(0, step):
        dttau.append(0)
        T.append(240)
        Tf.append(240)
        t.append(0)
        delt.append(0)
        tau.append(0)
        dttau.append(0)
        delT.append(0)
    T[1] = xdata[1]
    Tf[1] = T[1]
    t[1] = 0
    delt[1] = 0
    tau[1] = tau_0 * np.exp((xx*delhR / T[1]) + ((1 - xx)*delhR / Tf[1])-delhR/T_g)
    dttau[1] = delt[1] / tau[1]
    delT[1] = 0

    Tfinit = T[1]
    T_f_ns = [coordsLq[1][0]]
    for i in range(1, step):
        T[i] = xdata[i]
        delT[i] = xdata[i] - xdata[i-1]
        delt[i] = abs(delT[i]) / (rate/60)
        t[i] = t[i-1] + delt[i]
        tau[i] = tau_0 * np.exp((delhR*xx / T[i]) + ((1 - xx)*delhR / Tf[i-1])-delhR/T_g)
        for j in np.arange(2, i).reshape(-1):
            dttau[j] = dttau[j] + (delt[i] / tau[i])
            Tfinit = Tfinit + (delT[j] * (1 - np.exp(- (dttau[j] ** beta))))
            Tf[i] = Tfinit
        T_f_ns.append(Tfinit)
        Tfinit = Tf[1]

    print(T_f_ns[:10])
    return T_f_ns


####################################################
# Use T_g = T_fictive for tau determination
T_g = T_f_0[0]
# or plug in T_half, T_onset, ... from separate evaluation

# Start values
p0 = [10, 0.5, 0.4, 225085]


def fitTNMH(p, x):
    dTNMHdT = []
    modelTemp = np.append(x[::-1], x[1:])
    tau_0, xx, beta, deltaE = p
    TNMH = tnmfunc2(modelTemp, tau_0, xx, beta, deltaE)
    for i in range(len(modelTemp)-1):
        dTNMHdT.append((TNMH[i+1]-TNMH[i]) / (modelTemp[i+1]-modelTemp[i]))
    res = dTNMHdT[len(modelTemp)//2-1:]
    return np.array(res)  


# Use only every 20th data point to reduce fitting time
temperaturesInterpol = np.linspace(temperaturesCut[0], temperaturesCut[len(temperaturesCut)-1], 200)
dTfdTInterpol = np.interp(temperaturesInterpol, temperaturesCut, dTfdT)

ts = time.time()
model = odr.Model(fitTNMH)
whichFit = [1, 1, 1, 1]
odrdata = odr.Data(temperaturesInterpol, dTfdTInterpol)
myodr = odr.ODR(odrdata, model, beta0=p0, ifixb=whichFit, maxit=3000)
result = (myodr.run())
print(result.stopreason)
p0 = result.beta
print('TNMH heating Fit:')
print(p0)
# print('errors:')
# print(result.sd_beta)
ts = time.time()-ts
print('Time for fit:')
print(ts)
tau_0, xx, beta, deltaE = p0

fitResult = fitTNMH(p0, temperaturesInterpol)

"""
fig, (ax1) = P.subplots(1)
ax1.set_xlabel('Temperature / K')
ax1.set_ylabel('Fictive Temperature / K')
#ax1.set_xlim(left=coordsGl[0][0])
#ax1.set_xlim(right=coordsLq[1][0])
#ax1.set_ylim(bottom=coordsGl[0][1]-0.1)
#ax1.set_ylim(top=coordsLq[1][1]+0.1)
#ax1.plot(temperatures,fictiveTemp, 'bo')
#ax1.plot(temperatures,T_f_ns, 'r-')
ax1.plot(temperaturesCut,dTfdT, 'bo')
#ax1.plot(temperatures[0:5454],dTNMHdT[0:5454], 'bo')
#ax1.plot(temperatures[5456:],dTNMHdT[5455:], 'ro')
#ax1.plot(temperaturesModel[0:99],dTNMHdT[0:99], 'b-')
#ax1.plot(temperaturesModel[100:199],dTNMHdT[100:199], 'r-')
ax1.plot(temperaturesInterpol,dTfdTInterpol, 'yo')
ax1.plot(temperaturesInterpol,fitResult, 'r-')
P.show()
"""
dataOutName = os.path.splitext(str(dataName))[0] + "_dTfdT_tau.fit"
dataOut = np.column_stack((temperaturesInterpol, dTfdTInterpol, fitResult))

save = 0  # input("Save parameter to file? (yes = 1, no = 0):  ")
if save == '1':
    np.savetxt(dataOutName, dataOut, fmt='%1.5e', header="tau_TNMH="+str(tau_0) + "\tbeta="+str(beta) + "\tx="+str(xx) + "\tdeltaE="+str(deltaE) + "\nTemp\t\tdTfdT_data\t\tdTfdT_fit")
    # Try to determine the heating rate from the data file name (Searching for digits followed by ".")
    paramName = ((str(dataName)).split("_"))[0] + "_DSC_tau.param"
    print(paramName)
    with open(paramName, 'a') as f:
        f.write('\n#Heating Rate \t tau_TNMH \t beta \t\t\t x \t deltaE \t Tg fictive')
        f.write('\n')
        f.write('%1.4e \t %1.4e \t %1.4e \t %1.4e \t %1.4e \t %1.4e' % (rate, tau_0, beta, xx, deltaE, T_g))

    
